from zarr.core.buffer.cpu import (
    Buffer,
    NDBuffer,
    as_numpy_array_wrapper,
    buffer_prototype,
    numpy_buffer_prototype,
)

__all__ = [
    "Buffer",
    "NDBuffer",
    "as_numpy_array_wrapper",
    "buffer_prototype",
    "numpy_buffer_prototype",
]



from __future__ import annotations

import numpy as np

from tests.test_dtype.test_wrapper import BaseTestZDType
from zarr.core.dtype.npy.int import Int8, Int16, Int32, Int64, UInt8, UInt16, UInt32, UInt64


class TestInt8(BaseTestZDType):
    test_cls = Int8
    scalar_type = np.int8
    valid_dtype = (np.dtype(np.int8),)
    invalid_dtype = (
        np.dtype(np.int16),
        np.dtype(np.uint16),
        np.dtype(np.float64),
    )
    valid_json_v2 = ({"name": "|i1", "object_codec_id": None},)
    valid_json_v3 = ("int8",)
    invalid_json_v2 = (
        ">i1",
        "int8",
        "|f8",
    )
    invalid_json_v3 = (
        "|i1",
        "|f8",
        {"name": "int8", "configuration": {"endianness": "little"}},
    )

    scalar_v2_params = ((Int8(), 1), (Int8(), -1))
    scalar_v3_params = ((Int8(), 1), (Int8(), -1))
    cast_value_params = (
        (Int8(), 1, np.int8(1)),
        (Int8(), -1, np.int8(-1)),
    )
    invalid_scalar_params = ((Int8(), {"set!"}), (Int8(), ("tuple",)))
    item_size_params = (Int8(),)


class TestInt16(BaseTestZDType):
    test_cls = Int16
    scalar_type = np.int16
    valid_dtype = (np.dtype(">i2"), np.dtype("<i2"))
    invalid_dtype = (
        np.dtype(np.int8),
        np.dtype(np.uint16),
        np.dtype(np.float64),
    )
    valid_json_v2 = (
        {"name": ">i2", "object_codec_id": None},
        {"name": "<i2", "object_codec_id": None},
    )
    valid_json_v3 = ("int16",)
    invalid_json_v2 = (
        "|i2",
        "int16",
        "|f8",
    )
    invalid_json_v3 = (
        "|i2",
        "|f8",
        {"name": "int16", "configuration": {"endianness": "little"}},
    )

    scalar_v2_params = ((Int16(), 1), (Int16(), -1))
    scalar_v3_params = ((Int16(), 1), (Int16(), -1))
    cast_value_params = (
        (Int16(), 1, np.int16(1)),
        (Int16(), -1, np.int16(-1)),
    )
    invalid_scalar_params = ((Int16(), {"set!"}), (Int16(), ("tuple",)))
    item_size_params = (Int16(),)


class TestInt32(BaseTestZDType):
    test_cls = Int32
    scalar_type = np.int32
    # The behavior of some tests associated with this class variable are
    # order-dependent -- np.dtype('i') correctly fails certain tests only if it's not
    # in the last position of the tuple. I have no idea how this is possible!
    valid_dtype = (np.dtype("i"), np.dtype(">i4"), np.dtype("<i4"))
    invalid_dtype = (
        np.dtype(np.int8),
        np.dtype(np.uint16),
        np.dtype(np.float64),
    )
    valid_json_v2 = (
        {"name": ">i4", "object_codec_id": None},
        {"name": "<i4", "object_codec_id": None},
    )
    valid_json_v3 = ("int32",)
    invalid_json_v2 = (
        "|i4",
        "int32",
        "|f8",
    )
    invalid_json_v3 = (
        "|i4",
        "|f8",
        {"name": "int32", "configuration": {"endianness": "little"}},
    )

    scalar_v2_params = ((Int32(), 1), (Int32(), -1))
    scalar_v3_params = ((Int32(), 1), (Int32(), -1))
    cast_value_params = (
        (Int32(), 1, np.int32(1)),
        (Int32(), -1, np.int32(-1)),
    )
    invalid_scalar_params = ((Int32(), {"set!"}), (Int32(), ("tuple",)))
    item_size_params = (Int32(),)


class TestInt64(BaseTestZDType):
    test_cls = Int64
    scalar_type = np.int64
    valid_dtype = (np.dtype(">i8"), np.dtype("<i8"))
    invalid_dtype = (
        np.dtype(np.int8),
        np.dtype(np.uint16),
        np.dtype(np.float64),
    )
    valid_json_v2 = (
        {"name": ">i8", "object_codec_id": None},
        {"name": "<i8", "object_codec_id": None},
    )
    valid_json_v3 = ("int64",)
    invalid_json_v2 = (
        "|i8",
        "int64",
        "|f8",
    )
    invalid_json_v3 = (
        "|i8",
        "|f8",
        {"name": "int64", "configuration": {"endianness": "little"}},
    )

    scalar_v2_params = ((Int64(), 1), (Int64(), -1))
    scalar_v3_params = ((Int64(), 1), (Int64(), -1))
    cast_value_params = (
        (Int64(), 1, np.int64(1)),
        (Int64(), -1, np.int64(-1)),
    )
    invalid_scalar_params = ((Int64(), {"set!"}), (Int64(), ("tuple",)))
    item_size_params = (Int64(),)


class TestUInt8(BaseTestZDType):
    test_cls = UInt8
    scalar_type = np.uint8
    valid_dtype = (np.dtype(np.uint8),)
    invalid_dtype = (
        np.dtype(np.int8),
        np.dtype(np.int16),
        np.dtype(np.float64),
    )
    valid_json_v2 = ({"name": "|u1", "object_codec_id": None},)
    valid_json_v3 = ("uint8",)
    invalid_json_v2 = (
        "|u1",
        "uint8",
        "|f8",
    )
    invalid_json_v3 = (
        "|u1",
        "|f8",
        {"name": "uint8", "configuration": {"endianness": "little"}},
    )

    scalar_v2_params = ((UInt8(), 1), (UInt8(), 0))
    scalar_v3_params = ((UInt8(), 1), (UInt8(), 0))
    cast_value_params = (
        (UInt8(), 1, np.uint8(1)),
        (UInt8(), 0, np.uint8(0)),
    )
    invalid_scalar_params = ((UInt8(), {"set!"}), (UInt8(), ("tuple",)))
    item_size_params = (UInt8(),)


class TestUInt16(BaseTestZDType):
    test_cls = UInt16
    scalar_type = np.uint16
    valid_dtype = (np.dtype(">u2"), np.dtype("<u2"))
    invalid_dtype = (
        np.dtype(np.int8),
        np.dtype(np.int16),
        np.dtype(np.float64),
    )
    valid_json_v2 = (
        {"name": ">u2", "object_codec_id": None},
        {"name": "<u2", "object_codec_id": None},
    )
    valid_json_v3 = ("uint16",)
    invalid_json_v2 = (
        "|u2",
        "uint16",
        "|f8",
    )
    invalid_json_v3 = (
        "|u2",
        "|f8",
        {"name": "uint16", "configuration": {"endianness": "little"}},
    )

    scalar_v2_params = ((UInt16(), 1), (UInt16(), 0))
    scalar_v3_params = ((UInt16(), 1), (UInt16(), 0))
    cast_value_params = (
        (UInt16(), 1, np.uint16(1)),
        (UInt16(), 0, np.uint16(0)),
    )
    invalid_scalar_params = ((UInt16(), {"set!"}), (UInt16(), ("tuple",)))
    item_size_params = (UInt16(),)


class TestUInt32(BaseTestZDType):
    test_cls = UInt32
    scalar_type = np.uint32
    valid_dtype = (np.dtype(">u4"), np.dtype("<u4"))
    invalid_dtype = (
        np.dtype(np.int8),
        np.dtype(np.int16),
        np.dtype(np.float64),
    )
    valid_json_v2 = (
        {"name": ">u4", "object_codec_id": None},
        {"name": "<u4", "object_codec_id": None},
    )
    valid_json_v3 = ("uint32",)
    invalid_json_v2 = (
        "|u4",
        "uint32",
        "|f8",
    )
    invalid_json_v3 = (
        "|u4",
        "|f8",
        {"name": "uint32", "configuration": {"endianness": "little"}},
    )

    scalar_v2_params = ((UInt32(), 1), (UInt32(), 0))
    scalar_v3_params = ((UInt32(), 1), (UInt32(), 0))
    cast_value_params = (
        (UInt32(), 1, np.uint32(1)),
        (UInt32(), 0, np.uint32(0)),
    )
    invalid_scalar_params = ((UInt32(), {"set!"}), (UInt32(), ("tuple",)))
    item_size_params = (UInt32(),)


class TestUInt64(BaseTestZDType):
    test_cls = UInt64
    scalar_type = np.uint64
    valid_dtype = (np.dtype(">u8"), np.dtype("<u8"))
    invalid_dtype = (
        np.dtype(np.int8),
        np.dtype(np.int16),
        np.dtype(np.float64),
    )
    valid_json_v2 = (
        {"name": ">u8", "object_codec_id": None},
        {"name": "<u8", "object_codec_id": None},
    )
    valid_json_v3 = ("uint64",)
    invalid_json_v2 = (
        "|u8",
        "uint64",
        "|f8",
    )
    invalid_json_v3 = (
        "|u8",
        "|f8",
        {"name": "uint64", "configuration": {"endianness": "little"}},
    )

    scalar_v2_params = ((UInt64(), 1), (UInt64(), 0))
    scalar_v3_params = ((UInt64(), 1), (UInt64(), 0))
    cast_value_params = (
        (UInt64(), 1, np.uint64(1)),
        (UInt64(), 0, np.uint64(0)),
    )
    invalid_scalar_params = ((UInt64(), {"set!"}), (UInt64(), ("tuple",)))
    item_size_params = (UInt64(),)



from __future__ import annotations

import math

import numpy as np

from tests.test_dtype.test_wrapper import BaseTestZDType
from zarr.core.dtype.npy.complex import Complex64, Complex128


class _BaseTestFloat(BaseTestZDType):
    def scalar_equals(self, scalar1: object, scalar2: object) -> bool:
        if np.isnan(scalar1) and np.isnan(scalar2):  # type: ignore[call-overload]
            return True
        return super().scalar_equals(scalar1, scalar2)


class TestComplex64(_BaseTestFloat):
    test_cls = Complex64
    valid_dtype = (np.dtype(">c8"), np.dtype("<c8"))
    invalid_dtype = (
        np.dtype(np.int8),
        np.dtype(np.float64),
        np.dtype(np.complex128),
    )
    valid_json_v2 = (
        {"name": ">c8", "object_codec_id": None},
        {"name": "<c8", "object_codec_id": None},
    )
    valid_json_v3 = ("complex64",)
    invalid_json_v2 = (
        "|c8",
        "complex64",
        "|f8",
    )
    invalid_json_v3 = (
        "|c8",
        "|f8",
        {"name": "complex64", "configuration": {"endianness": "little"}},
    )

    scalar_v2_params = (
        (Complex64(), (1.0, 1.0)),
        (Complex64(), (-1.0, "Infinity")),
        (Complex64(), (0, "NaN")),
    )
    scalar_v3_params = (
        (Complex64(), (1.0, 1.0)),
        (Complex64(), (-1.0, "Infinity")),
        (Complex64(), (0, "NaN")),
    )
    cast_value_params = (
        (Complex64(), complex(1.0, 1.0), np.complex64(complex(1.0, 1.0))),
        (Complex64(), complex(-1.0, math.inf), np.complex64(complex(-1.0, math.inf))),
        (Complex64(), complex(0, math.nan), np.complex64(complex(0, math.nan))),
    )
    invalid_scalar_params = ((Complex64(), {"type": "dict"}),)
    item_size_params = (Complex64(),)


class TestComplex128(_BaseTestFloat):
    test_cls = Complex128
    valid_dtype = (np.dtype(">c16"), np.dtype("<c16"))
    invalid_dtype = (
        np.dtype(np.int8),
        np.dtype(np.float64),
        np.dtype(np.complex64),
    )
    valid_json_v2 = (
        {"name": ">c16", "object_codec_id": None},
        {"name": "<c16", "object_codec_id": None},
    )
    valid_json_v3 = ("complex128",)
    invalid_json_v2 = (
        "|c16",
        "complex128",
        "|f8",
    )
    invalid_json_v3 = (
        "|c16",
        "|f8",
        {"name": "complex128", "configuration": {"endianness": "little"}},
    )

    scalar_v2_params = (
        (Complex128(), (1.0, 1.0)),
        (Complex128(), (-1.0, "Infinity")),
        (Complex128(), (0, "NaN")),
    )
    scalar_v3_params = (
        (Complex128(), (1.0, 1.0)),
        (Complex128(), (-1.0, "Infinity")),
        (Complex128(), (0, "NaN")),
    )
    cast_value_params = (
        (Complex128(), complex(1.0, 1.0), np.complex128(complex(1.0, 1.0))),
        (Complex128(), complex(-1.0, math.inf), np.complex128(complex(-1.0, math.inf))),
        (Complex128(), complex(0, math.nan), np.complex128(complex(0, math.nan))),
    )
    invalid_scalar_params = ((Complex128(), {"type": "dict"}),)
    item_size_params = (Complex128(),)



from __future__ import annotations

import contextlib
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Self

import numpy as np

from zarr.core.dtype.common import (
    DataTypeValidationError,
    DTypeJSON,
)

if TYPE_CHECKING:
    from importlib.metadata import EntryPoint

    from zarr.core.common import ZarrFormat
    from zarr.core.dtype.wrapper import TBaseDType, TBaseScalar, ZDType


# This class is different from the other registry classes, which inherit from
# dict. IMO it's simpler to just do a dataclass. But long-term we should
# have just 1 registry class in use.
@dataclass(frozen=True, kw_only=True)
class DataTypeRegistry:
    """
    A registry for ZDType classes.

    This registry is a mapping from Zarr data type names to their
    corresponding ZDType classes.

    Attributes
    ----------
    contents : dict[str, type[ZDType[TBaseDType, TBaseScalar]]]
        The mapping from Zarr data type names to their corresponding
        ZDType classes.
    """

    contents: dict[str, type[ZDType[TBaseDType, TBaseScalar]]] = field(
        default_factory=dict, init=False
    )

    _lazy_load_list: list[EntryPoint] = field(default_factory=list, init=False)

    def _lazy_load(self) -> None:
        """
        Load all data types from the lazy load list and register them with
        the registry. After loading, clear the lazy load list.
        """
        for e in self._lazy_load_list:
            self.register(e.load()._zarr_v3_name, e.load())

        self._lazy_load_list.clear()

    def register(self: Self, key: str, cls: type[ZDType[TBaseDType, TBaseScalar]]) -> None:
        """
        Register a data type with the registry.

        Parameters
        ----------
        key : str
            The Zarr V3 name of the data type.
        cls : type[ZDType[TBaseDType, TBaseScalar]]
            The class of the data type to register.

        Notes
        -----
        This method is idempotent. If the data type is already registered, this
        method does nothing.
        """
        if key not in self.contents or self.contents[key] != cls:
            self.contents[key] = cls

    def unregister(self, key: str) -> None:
        """
        Unregister a data type from the registry.

        Parameters
        ----------
        key : str
            The key associated with the ZDType class to be unregistered.

        Returns
        -------
        None

        Raises
        ------
        KeyError
            If the data type is not found in the registry.
        """
        if key in self.contents:
            del self.contents[key]
        else:
            raise KeyError(f"Data type '{key}' not found in registry.")

    def get(self, key: str) -> type[ZDType[TBaseDType, TBaseScalar]]:
        """
        Retrieve a registered ZDType class by its key.

        Parameters
        ----------
        key : str
            The key associated with the desired ZDType class.

        Returns
        -------
        type[ZDType[TBaseDType, TBaseScalar]]
            The ZDType class registered under the given key.

        Raises
        ------
        KeyError
            If the key is not found in the registry.
        """

        return self.contents[key]

    def match_dtype(self, dtype: TBaseDType) -> ZDType[TBaseDType, TBaseScalar]:
        """
        Match a native data type, e.g. a NumPy data type, to a registered ZDType.

        Parameters
        ----------
        dtype : TBaseDType
            The native data type to match.

        Returns
        -------
        ZDType[TBaseDType, TBaseScalar]
            The matched ZDType corresponding to the provided NumPy data type.

        Raises
        ------
        ValueError
            If the data type is a NumPy "Object" type, which is ambiguous, or if multiple
            or no Zarr data types are found that match the provided dtype.

        Notes
        -----
        This function attempts to resolve a Zarr data type from a given native data type.
        If the dtype is a NumPy "Object" data type, it raises a ValueError, as this type
        can represent multiple Zarr data types. In such cases, a specific Zarr data type
        should be explicitly constructed instead of relying on dynamic resolution.

        If multiple matches are found, it will also raise a ValueError. In this case
        conflicting data types must be unregistered, or the Zarr data type should be explicitly
        constructed.
        """

        if dtype == np.dtype("O"):
            msg = (
                f"Zarr data type resolution from {dtype} failed. "
                'Attempted to resolve a zarr data type from a numpy "Object" data type, which is '
                'ambiguous, as multiple zarr data types can be represented by the numpy "Object" '
                "data type. "
                "In this case you should construct your array by providing a specific Zarr data "
                'type. For a list of Zarr data types that are compatible with the numpy "Object"'
                "data type, see https://github.com/zarr-developers/zarr-python/issues/3117"
            )
            raise ValueError(msg)
        matched: list[ZDType[TBaseDType, TBaseScalar]] = []
        for val in self.contents.values():
            with contextlib.suppress(DataTypeValidationError):
                matched.append(val.from_native_dtype(dtype))
        if len(matched) == 1:
            return matched[0]
        elif len(matched) > 1:
            msg = (
                f"Zarr data type resolution from {dtype} failed. "
                f"Multiple data type wrappers found that match dtype '{dtype}': {matched}. "
                "You should unregister one of these data types, or avoid Zarr data type inference "
                "entirely by providing a specific Zarr data type when creating your array."
                "For more information, see https://github.com/zarr-developers/zarr-python/issues/3117"
            )
            raise ValueError(msg)
        raise ValueError(f"No Zarr data type found that matches dtype '{dtype!r}'")

    def match_json(
        self, data: DTypeJSON, *, zarr_format: ZarrFormat
    ) -> ZDType[TBaseDType, TBaseScalar]:
        """
        Match a JSON representation of a data type to a registered ZDType.

        Parameters
        ----------
        data : DTypeJSON
            The JSON representation of a data type to match.
        zarr_format : ZarrFormat
            The Zarr format version to consider when matching data types.

        Returns
        -------
        ZDType[TBaseDType, TBaseScalar]
            The matched ZDType corresponding to the JSON representation.

        Raises
        ------
        ValueError
            If no matching Zarr data type is found for the given JSON data.
        """

        for val in self.contents.values():
            try:
                return val.from_json(data, zarr_format=zarr_format)
            except DataTypeValidationError:
                pass
        raise ValueError(f"No Zarr data type found that matches {data!r}")



# /// script
# requires-python = "==3.12"
# dependencies = [
#  "zarr==3.0.8"
# ]
# ///


import argparse

import zarr
from zarr.abc.store import Store

def copy_group(
    *, node: zarr.Group, store: Store, path: str, overwrite: bool
) -> zarr.Group:
    result = zarr.create_group(
        store=store,
        path=path,
        overwrite=overwrite,
        attributes=node.attrs.asdict(),
        zarr_format=node.metadata.zarr_format)
    for key, child in node.members():
        child_path = f"{path}/{key}"
        if isinstance(child, zarr.Group):
            copy_group(node=child, store=store, path=child_path, overwrite=overwrite)
        else:
            copy_array(node=child, store=store, overwrite=overwrite, path=child_path)
    return result


def copy_array(
    *, node: zarr.Array, store: Store, path: str, overwrite: bool
) -> zarr.Array:
    result = zarr.from_array(store, name=path, data=node, write_data=True)
    return result


def copy_node(
    node: zarr.Group | zarr.Array, store: Store, path: str, overwrite: bool
) -> zarr.Group | zarr.Array:
    if isinstance(node, zarr.Group):
        return copy_group(node=node, store=store, path=path, overwrite=overwrite)
    else:
        return copy_array(node=node, store=store, path=path, overwrite=overwrite)


def cli() -> None:
    parser = argparse.ArgumentParser(
        description="Copy a zarr hierarchy from one location to another"
    )
    parser.add_argument("source", type=str, help="Path to the source zarr hierarchy")
    parser.add_argument("destination", type=str, help="Path to the destination zarr hierarchy")
    args = parser.parse_args()

    src, dst = args.source, args.destination
    root_src = zarr.open(src, mode="r")
    result = copy_node(node=root_src, store=dst, path="", overwrite=True)

    print(f"successfully created {result} at {dst}")


def main() -> None:
    cli()


if __name__ == "__main__":
    main()



from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Final, TypeAlias

from zarr.core.dtype.common import (
    DataTypeValidationError,
    DTypeJSON,
)
from zarr.core.dtype.npy.bool import Bool
from zarr.core.dtype.npy.bytes import (
    NullTerminatedBytes,
    NullterminatedBytesJSON_V2,
    NullTerminatedBytesJSON_V3,
    RawBytes,
    RawBytesJSON_V2,
    RawBytesJSON_V3,
    VariableLengthBytes,
    VariableLengthBytesJSON_V2,
)
from zarr.core.dtype.npy.complex import Complex64, Complex128
from zarr.core.dtype.npy.float import Float16, Float32, Float64
from zarr.core.dtype.npy.int import Int8, Int16, Int32, Int64, UInt8, UInt16, UInt32, UInt64
from zarr.core.dtype.npy.structured import Structured, StructuredJSON_V2, StructuredJSON_V3
from zarr.core.dtype.npy.time import (
    DateTime64,
    DateTime64JSON_V2,
    DateTime64JSON_V3,
    TimeDelta64,
    TimeDelta64JSON_V2,
    TimeDelta64JSON_V3,
)

if TYPE_CHECKING:
    from zarr.core.common import ZarrFormat

from collections.abc import Mapping

import numpy as np
import numpy.typing as npt

from zarr.core.common import JSON
from zarr.core.dtype.npy.string import (
    FixedLengthUTF32,
    FixedLengthUTF32JSON_V2,
    FixedLengthUTF32JSON_V3,
    VariableLengthUTF8,
    VariableLengthUTF8JSON_V2,
)
from zarr.core.dtype.registry import DataTypeRegistry
from zarr.core.dtype.wrapper import TBaseDType, TBaseScalar, ZDType

__all__ = [
    "Bool",
    "Complex64",
    "Complex128",
    "DataTypeRegistry",
    "DataTypeValidationError",
    "DateTime64",
    "DateTime64JSON_V2",
    "DateTime64JSON_V3",
    "FixedLengthUTF32",
    "FixedLengthUTF32JSON_V2",
    "FixedLengthUTF32JSON_V3",
    "Float16",
    "Float32",
    "Float64",
    "Int8",
    "Int16",
    "Int32",
    "Int64",
    "NullTerminatedBytes",
    "NullTerminatedBytesJSON_V3",
    "NullterminatedBytesJSON_V2",
    "RawBytes",
    "RawBytesJSON_V2",
    "RawBytesJSON_V3",
    "Structured",
    "StructuredJSON_V2",
    "StructuredJSON_V3",
    "TBaseDType",
    "TBaseScalar",
    "TimeDelta64",
    "TimeDelta64",
    "TimeDelta64JSON_V2",
    "TimeDelta64JSON_V3",
    "UInt8",
    "UInt16",
    "UInt32",
    "UInt64",
    "VariableLengthBytes",
    "VariableLengthBytesJSON_V2",
    "VariableLengthUTF8",
    "VariableLengthUTF8JSON_V2",
    "ZDType",
    "data_type_registry",
    "parse_data_type",
    "parse_dtype",
]

data_type_registry = DataTypeRegistry()

IntegerDType = Int8 | Int16 | Int32 | Int64 | UInt8 | UInt16 | UInt32 | UInt64
INTEGER_DTYPE: Final = Int8, Int16, Int32, Int64, UInt8, UInt16, UInt32, UInt64

FloatDType = Float16 | Float32 | Float64
FLOAT_DTYPE: Final = Float16, Float32, Float64

ComplexFloatDType = Complex64 | Complex128
COMPLEX_FLOAT_DTYPE: Final = Complex64, Complex128

StringDType = FixedLengthUTF32 | VariableLengthUTF8
STRING_DTYPE: Final = FixedLengthUTF32, VariableLengthUTF8

TimeDType = DateTime64 | TimeDelta64
TIME_DTYPE: Final = DateTime64, TimeDelta64

BytesDType = RawBytes | NullTerminatedBytes | VariableLengthBytes
BYTES_DTYPE: Final = RawBytes, NullTerminatedBytes, VariableLengthBytes

AnyDType = (
    Bool
    | IntegerDType
    | FloatDType
    | ComplexFloatDType
    | StringDType
    | BytesDType
    | Structured
    | TimeDType
    | VariableLengthBytes
)
# mypy has trouble inferring the type of variablelengthstring dtype, because its class definition
# depends on the installed numpy version. That's why the type: ignore statement is needed here.
ANY_DTYPE: Final = (
    Bool,
    *INTEGER_DTYPE,
    *FLOAT_DTYPE,
    *COMPLEX_FLOAT_DTYPE,
    *STRING_DTYPE,
    *BYTES_DTYPE,
    Structured,
    *TIME_DTYPE,
    VariableLengthBytes,
)

# These are aliases for variable-length UTF-8 strings
# We handle them when a user requests a data type instead of using NumPy's dtype inferece because
# the default NumPy behavior -- to inspect the user-provided array data and choose
# an appropriately sized U dtype -- is unworkable for Zarr.
VLEN_UTF8_ALIAS: Final = ("str", str, "string")

# This type models inputs that can be coerced to a ZDType
ZDTypeLike: TypeAlias = npt.DTypeLike | ZDType[TBaseDType, TBaseScalar] | Mapping[str, JSON] | str

for dtype in ANY_DTYPE:
    # mypy does not know that all the elements of ANY_DTYPE are subclasses of ZDType
    data_type_registry.register(dtype._zarr_v3_name, dtype)  # type: ignore[arg-type]


# TODO: find a better name for this function
def get_data_type_from_native_dtype(dtype: npt.DTypeLike) -> ZDType[TBaseDType, TBaseScalar]:
    """
    Get a data type wrapper (an instance of ``ZDType``) from a native data type, e.g. a numpy dtype.
    """
    if not isinstance(dtype, np.dtype):
        na_dtype: np.dtype[np.generic]
        if isinstance(dtype, list):
            # this is a valid _VoidDTypeLike check
            na_dtype = np.dtype([tuple(d) for d in dtype])
        else:
            na_dtype = np.dtype(dtype)
    else:
        na_dtype = dtype
    return data_type_registry.match_dtype(dtype=na_dtype)


def get_data_type_from_json(
    dtype_spec: DTypeJSON, *, zarr_format: ZarrFormat
) -> ZDType[TBaseDType, TBaseScalar]:
    """
    Given a JSON representation of a data type and a Zarr format version,
    attempt to create a ZDType instance from the registered ZDType classes.
    """
    return data_type_registry.match_json(dtype_spec, zarr_format=zarr_format)


def parse_data_type(
    dtype_spec: ZDTypeLike,
    *,
    zarr_format: ZarrFormat,
) -> ZDType[TBaseDType, TBaseScalar]:
    """
    Interpret the input as a ZDType.

    This function wraps ``parse_dtype``. The only difference is the function name. This function may
    be deprecated in a future version of Zarr Python in favor of ``parse_dtype``.

    Parameters
    ----------
    dtype_spec : ZDTypeLike
        The input to be interpreted as a ZDType. This could be a ZDType, which will be returned
        directly, or a JSON representation of a ZDType, or a native dtype, or a python object that
        can be converted into a native dtype.
    zarr_format : ZarrFormat
        The Zarr format version. This parameter is required because this function will attempt to
        parse the JSON representation of a data type, and the JSON representation of data types
        varies between Zarr 2 and Zarr 3.

    Returns
    -------
    ZDType[TBaseDType, TBaseScalar]
        The ZDType corresponding to the input.

    Examples
    --------
    >>> from zarr.dtype import parse_data_type
    >>> import numpy as np
    >>> parse_data_type("int32", zarr_format=2)
    Int32(endianness='little')
    >>> parse_data_type(np.dtype('S10'), zarr_format=2)
    NullTerminatedBytes(length=10)
    >>> parse_data_type({"name": "numpy.datetime64", "configuration": {"unit": "s", "scale_factor": 10}}, zarr_format=3)
    DateTime64(endianness='little', scale_factor=10, unit='s')
    """
    return parse_dtype(dtype_spec, zarr_format=zarr_format)


def parse_dtype(
    dtype_spec: ZDTypeLike,
    *,
    zarr_format: ZarrFormat,
) -> ZDType[TBaseDType, TBaseScalar]:
    """
    Convert the input as a ZDType.

    Parameters
    ----------
    dtype_spec : ZDTypeLike
        The input to be converted to a ZDType. This could be a ZDType, which will be returned
        directly, or a JSON representation of a ZDType, or a numpy dtype, or a python object that
        can be converted into a native dtype.
    zarr_format : ZarrFormat
        The Zarr format version. This parameter is required because this function will attempt to
        parse the JSON representation of a data type, and the JSON representation of data types
        varies between Zarr 2 and Zarr 3.

    Returns
    -------
    ZDType[TBaseDType, TBaseScalar]
        The ZDType corresponding to the input.

    Examples
    --------
    >>> from zarr.dtype import parse_dtype
    >>> import numpy as np
    >>> parse_dtype("int32", zarr_format=2)
    Int32(endianness='little')
    >>> parse_dtype(np.dtype('S10'), zarr_format=2)
    NullTerminatedBytes(length=10)
    >>> parse_dtype({"name": "numpy.datetime64", "configuration": {"unit": "s", "scale_factor": 10}}, zarr_format=3)
    DateTime64(endianness='little', scale_factor=10, unit='s')
    """
    if isinstance(dtype_spec, ZDType):
        return dtype_spec
    # First attempt to interpret the input as JSON
    if isinstance(dtype_spec, Mapping | str | Sequence):
        try:
            return get_data_type_from_json(dtype_spec, zarr_format=zarr_format)  # type: ignore[arg-type]
        except ValueError:
            # no data type matched this JSON-like input
            pass
    if dtype_spec in VLEN_UTF8_ALIAS:
        # If the dtype request is one of the aliases for variable-length UTF-8 strings,
        # return that dtype.
        return VariableLengthUTF8()  # type: ignore[return-value]
    # otherwise, we have either a numpy dtype string, or a zarr v3 dtype string, and in either case
    # we can create a native dtype from it, and do the dtype inference from that
    return get_data_type_from_native_dtype(dtype_spec)  # type: ignore[arg-type]



from __future__ import annotations

from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    ClassVar,
    Literal,
    Self,
    TypeGuard,
    overload,
)

import numpy as np

from zarr.core.dtype.common import (
    DataTypeValidationError,
    DTypeConfig_V2,
    DTypeJSON,
    HasEndianness,
    HasItemSize,
    check_dtype_spec_v2,
)
from zarr.core.dtype.npy.common import (
    ComplexLike,
    TComplexDType_co,
    TComplexScalar_co,
    check_json_complex_float_v2,
    check_json_complex_float_v3,
    complex_float_from_json_v2,
    complex_float_from_json_v3,
    complex_float_to_json_v2,
    complex_float_to_json_v3,
    endianness_to_numpy_str,
    get_endianness_from_numpy_dtype,
)
from zarr.core.dtype.wrapper import TBaseDType, ZDType

if TYPE_CHECKING:
    from zarr.core.common import JSON, ZarrFormat


@dataclass(frozen=True)
class BaseComplex(ZDType[TComplexDType_co, TComplexScalar_co], HasEndianness, HasItemSize):
    """
    A base class for Zarr data types that wrap NumPy complex float data types.
    """

    # This attribute holds the possible zarr v2 JSON names for the data type
    _zarr_v2_names: ClassVar[tuple[str, ...]]

    @classmethod
    def from_native_dtype(cls, dtype: TBaseDType) -> Self:
        """
        Create an instance of this data type from a NumPy complex dtype.

        Parameters
        ----------
        dtype : TBaseDType
            The native dtype to convert.

        Returns
        -------
        Self
            An instance of this data type.

        Raises
        ------
        DataTypeValidationError
            If the dtype is not compatible with this data type.
        """
        if cls._check_native_dtype(dtype):
            return cls(endianness=get_endianness_from_numpy_dtype(dtype))
        raise DataTypeValidationError(
            f"Invalid data type: {dtype}. Expected an instance of {cls.dtype_cls}"
        )

    def to_native_dtype(self) -> TComplexDType_co:
        """
        Convert this class to a NumPy complex dtype with the appropriate byte order.

        Returns
        -------
        TComplexDType_co
            A NumPy data type object representing the complex data type with the specified byte order.
        """

        byte_order = endianness_to_numpy_str(self.endianness)
        return self.dtype_cls().newbyteorder(byte_order)  # type: ignore[return-value]

    @classmethod
    def _check_json_v2(cls, data: DTypeJSON) -> TypeGuard[DTypeConfig_V2[str, None]]:
        """
        Check that the input is a valid JSON representation of this data type.

        The input data must be a mapping that contains a "name" key that is one of
        the strings from cls._zarr_v2_names and an "object_codec_id" key that is None.

        Parameters
        ----------
        data : DTypeJSON
            The JSON data to check.

        Returns
        -------
        bool
            True if the input is a valid JSON representation, False otherwise.
        """
        return (
            check_dtype_spec_v2(data)
            and data["name"] in cls._zarr_v2_names
            and data["object_codec_id"] is None
        )

    @classmethod
    def _check_json_v3(cls, data: DTypeJSON) -> TypeGuard[str]:
        """
        Check that the input is a valid JSON representation of this data type in Zarr V3.

        This method verifies that the provided data matches the expected Zarr V3
        representation, which is the string specified by the class-level attribute _zarr_v3_name.

        Parameters
        ----------
        data : DTypeJSON
            The JSON data to check.

        Returns
        -------
        TypeGuard[str]
            True if the input is a valid representation of this class in Zarr V3, False otherwise.
        """

        return data == cls._zarr_v3_name

    @classmethod
    def _from_json_v2(cls, data: DTypeJSON) -> Self:
        """
        Create an instance of this class from Zarr V2-flavored JSON.

        Parameters
        ----------
        data : DTypeJSON
            The JSON data.

        Returns
        -------
        Self
            An instance of this class.

        Raises
        ------
        DataTypeValidationError
            If the input JSON is not a valid representation of this class.
        """
        if cls._check_json_v2(data):
            # Going via numpy ensures that we get the endianness correct without
            # annoying string parsing.
            name = data["name"]
            return cls.from_native_dtype(np.dtype(name))
        msg = f"Invalid JSON representation of {cls.__name__}. Got {data!r}, expected one of the strings {cls._zarr_v2_names}."
        raise DataTypeValidationError(msg)

    @classmethod
    def _from_json_v3(cls, data: DTypeJSON) -> Self:
        """
        Create an instance of this class from Zarr V3-flavored JSON.

        Parameters
        ----------
        data : DTypeJSON
            The JSON data.

        Returns
        -------
        Self
            An instance of this data type.

        Raises
        ------
        DataTypeValidationError
            If the input JSON is not a valid representation of this class.
        """
        if cls._check_json_v3(data):
            return cls()
        msg = f"Invalid JSON representation of {cls.__name__}. Got {data!r}, expected {cls._zarr_v3_name}."
        raise DataTypeValidationError(msg)

    @overload
    def to_json(self, zarr_format: Literal[2]) -> DTypeConfig_V2[str, None]: ...

    @overload
    def to_json(self, zarr_format: Literal[3]) -> str: ...

    def to_json(self, zarr_format: ZarrFormat) -> DTypeConfig_V2[str, None] | str:
        """
        Serialize this object to a JSON-serializable representation.

        Parameters
        ----------
        zarr_format : ZarrFormat
            The Zarr format version. Supported values are 2 and 3.

        Returns
        -------
        DTypeConfig_V2[str, None] | str
            If ``zarr_format`` is 2, a dictionary with ``"name"`` and ``"object_codec_id"`` keys is
            returned.
            If ``zarr_format`` is 3, a string representation of the complex data type is returned.

        Raises
        ------
        ValueError
            If `zarr_format` is not 2 or 3.
        """

        if zarr_format == 2:
            return {"name": self.to_native_dtype().str, "object_codec_id": None}
        elif zarr_format == 3:
            return self._zarr_v3_name
        raise ValueError(f"zarr_format must be 2 or 3, got {zarr_format}")  # pragma: no cover

    def _check_scalar(self, data: object) -> TypeGuard[ComplexLike]:
        """
        Check that the input is a scalar complex value.

        Parameters
        ----------
        data : object
            The value to check.

        Returns
        -------
        TypeGuard[ComplexLike]
            True if the input is a scalar complex value, False otherwise.
        """
        return isinstance(data, ComplexLike)

    def _cast_scalar_unchecked(self, data: ComplexLike) -> TComplexScalar_co:
        """
        Cast the provided scalar data to the native scalar type of this class.

        Parameters
        ----------
        data : ComplexLike
            The data to cast.

        Returns
        -------
        TComplexScalar_co
            The casted data as a numpy complex scalar.

        Notes
        -----
        This method does not perform any type checking.
        The input data must be a scalar complex value.
        """
        return self.to_native_dtype().type(data)  # type: ignore[return-value]

    def cast_scalar(self, data: object) -> TComplexScalar_co:
        """
        Attempt to cast a given object to a numpy complex scalar.

        Parameters
        ----------
        data : object
            The data to be cast to a numpy complex scalar.

        Returns
        -------
        TComplexScalar_co
            The data cast as a numpy complex scalar.

        Raises
        ------
        TypeError
            If the data cannot be converted to a numpy complex scalar.
        """
        if self._check_scalar(data):
            return self._cast_scalar_unchecked(data)
        msg = (
            f"Cannot convert object {data!r} with type {type(data)} to a scalar compatible with the "
            f"data type {self}."
        )
        raise TypeError(msg)

    def default_scalar(self) -> TComplexScalar_co:
        """
        Get the default value, which is 0 cast to this dtype

        Returns
        -------
        Int scalar
            The default value.
        """
        return self._cast_scalar_unchecked(0)

    def from_json_scalar(self, data: JSON, *, zarr_format: ZarrFormat) -> TComplexScalar_co:
        """
        Read a JSON-serializable value as a numpy float.

        Parameters
        ----------
        data : JSON
            The JSON-serializable value.
        zarr_format : ZarrFormat
            The zarr format version.

        Returns
        -------
        TScalar_co
            The numpy float.
        """
        if zarr_format == 2:
            if check_json_complex_float_v2(data):
                return self._cast_scalar_unchecked(complex_float_from_json_v2(data))
            raise TypeError(
                f"Invalid type: {data}. Expected a float or a special string encoding of a float."
            )
        elif zarr_format == 3:
            if check_json_complex_float_v3(data):
                return self._cast_scalar_unchecked(complex_float_from_json_v3(data))
            raise TypeError(
                f"Invalid type: {data}. Expected a float or a special string encoding of a float."
            )
        raise ValueError(f"zarr_format must be 2 or 3, got {zarr_format}")  # pragma: no cover

    def to_json_scalar(self, data: object, *, zarr_format: ZarrFormat) -> JSON:
        """
        Convert an object to a JSON-serializable float.

        Parameters
        ----------
        data : _BaseScalar
            The value to convert.
        zarr_format : ZarrFormat
            The zarr format version.

        Returns
        -------
        JSON
            The JSON-serializable form of the complex number, which is a list of two floats,
            each of which is encoding according to a zarr-format-specific encoding.
        """
        if zarr_format == 2:
            return complex_float_to_json_v2(self.cast_scalar(data))
        elif zarr_format == 3:
            return complex_float_to_json_v3(self.cast_scalar(data))
        raise ValueError(f"zarr_format must be 2 or 3, got {zarr_format}")  # pragma: no cover


@dataclass(frozen=True, kw_only=True)
class Complex64(BaseComplex[np.dtypes.Complex64DType, np.complex64]):
    """
    A Zarr data type for arrays containing 64 bit complex floats.

    Wraps the ``np.dtypes.Complex64DType`` data type. Scalars for this data type
    are instances of ``np.complex64``.

    Attributes
    ----------
    dtype_cls : Type[np.dtypes.Complex64DType]
        The numpy dtype class for this data type.
    _zarr_v3_name : ClassVar[Literal["complex64"]]
        The name of this data type in Zarr V3.
    _zarr_v2_names : ClassVar[tuple[Literal[">c8"], Literal["<c8"]]]
        The names of this data type in Zarr V2.
    """

    dtype_cls = np.dtypes.Complex64DType
    _zarr_v3_name: ClassVar[Literal["complex64"]] = "complex64"
    _zarr_v2_names: ClassVar[tuple[Literal[">c8"], Literal["<c8"]]] = (">c8", "<c8")

    @property
    def item_size(self) -> int:
        """
        The size of a single scalar in bytes.

        Returns
        -------
        int
            The size of a single scalar in bytes.
        """
        return 8


@dataclass(frozen=True, kw_only=True)
class Complex128(BaseComplex[np.dtypes.Complex128DType, np.complex128], HasEndianness):
    """
    A Zarr data type for arrays containing 64 bit complex floats.

    Wraps the ``np.dtypes.Complex128DType`` data type. Scalars for this data type
    are instances of ``np.complex128``.

    Attributes
    ----------
    dtype_cls : Type[np.dtypes.Complex128DType]
        The numpy dtype class for this data type.
    _zarr_v3_name : ClassVar[Literal["complex128"]]
        The name of this data type in Zarr V3.
    _zarr_v2_names : ClassVar[tuple[Literal[">c16"], Literal["<c16"]]]
        The names of this data type in Zarr V2.
    """

    dtype_cls = np.dtypes.Complex128DType
    _zarr_v3_name: ClassVar[Literal["complex128"]] = "complex128"
    _zarr_v2_names: ClassVar[tuple[Literal[">c16"], Literal["<c16"]]] = (">c16", "<c16")

    @property
    def item_size(self) -> int:
        """
        The size of a single scalar in bytes.

        Returns
        -------
        int
            The size of a single scalar in bytes.
        """
        return 16



import pytest


def pytest_configure(config: pytest.Config) -> None:
    # The tests in zarr.testing are intended to be run by downstream projects.
    # To allow those downstream projects to run with `--strict-markers`, we need
    # to register an entry point with pytest11 and register our "plugin" with it,
    # which just registers the markers used in zarr.testing
    config.addinivalue_line("markers", "gpu: mark a test as requiring CuPy and GPU")



from __future__ import annotations

import numpy as np

from tests.test_dtype.test_wrapper import BaseTestZDType
from zarr.core.dtype.npy.float import Float16, Float32, Float64


class _BaseTestFloat(BaseTestZDType):
    def scalar_equals(self, scalar1: object, scalar2: object) -> bool:
        if np.isnan(scalar1) and np.isnan(scalar2):  # type: ignore[call-overload]
            return True
        return super().scalar_equals(scalar1, scalar2)

    hex_string_params: tuple[tuple[str, float], ...] = ()

    def test_hex_encoding(self, hex_string_params: tuple[str, float]) -> None:
        """
        Test that hexadecimal strings can be read as NaN values
        """
        hex_string, expected = hex_string_params
        zdtype = self.test_cls()
        observed = zdtype.from_json_scalar(hex_string, zarr_format=3)
        assert self.scalar_equals(observed, expected)


class TestFloat16(_BaseTestFloat):
    test_cls = Float16
    valid_dtype = (np.dtype(">f2"), np.dtype("<f2"))
    invalid_dtype = (
        np.dtype(np.int8),
        np.dtype(np.uint16),
        np.dtype(np.float32),
    )
    valid_json_v2 = (
        {"name": ">f2", "object_codec_id": None},
        {"name": "<f2", "object_codec_id": None},
    )
    valid_json_v3 = ("float16",)
    invalid_json_v2 = (
        "|f2",
        "float16",
        "|i1",
    )
    invalid_json_v3 = (
        "|f2",
        "|i1",
        {"name": "float16", "configuration": {"endianness": "little"}},
    )

    scalar_v2_params = (
        (Float16(), 1.0),
        (Float16(), -1.0),
        (Float16(), "NaN"),
        (Float16(), "Infinity"),
    )
    scalar_v3_params = (
        (Float16(), 1.0),
        (Float16(), -1.0),
        (Float16(), "NaN"),
        (Float16(), "Infinity"),
    )
    cast_value_params = (
        (Float16(), 1.0, np.float16(1.0)),
        (Float16(), -1.0, np.float16(-1.0)),
        (Float16(), "NaN", np.float16("NaN")),
    )
    invalid_scalar_params = ((Float16(), {"set!"}),)
    hex_string_params = (("0x7fc0", np.nan), ("0x7fc1", np.nan), ("0x3c00", 1.0))
    item_size_params = (Float16(),)


class TestFloat32(_BaseTestFloat):
    test_cls = Float32
    scalar_type = np.float32
    valid_dtype = (np.dtype(">f4"), np.dtype("<f4"))
    invalid_dtype = (
        np.dtype(np.int8),
        np.dtype(np.uint16),
        np.dtype(np.float64),
    )
    valid_json_v2 = (
        {"name": ">f4", "object_codec_id": None},
        {"name": "<f4", "object_codec_id": None},
    )
    valid_json_v3 = ("float32",)
    invalid_json_v2 = (
        "|f4",
        "float32",
        "|i1",
    )
    invalid_json_v3 = (
        "|f4",
        "|i1",
        {"name": "float32", "configuration": {"endianness": "little"}},
    )

    scalar_v2_params = (
        (Float32(), 1.0),
        (Float32(), -1.0),
        (Float32(), "NaN"),
        (Float32(), "Infinity"),
    )
    scalar_v3_params = (
        (Float32(), 1.0),
        (Float32(), -1.0),
        (Float32(), "NaN"),
        (Float32(), "Infinity"),
    )

    cast_value_params = (
        (Float32(), 1.0, np.float32(1.0)),
        (Float32(), -1.0, np.float32(-1.0)),
        (Float32(), "NaN", np.float32("NaN")),
    )
    invalid_scalar_params = ((Float32(), {"set!"}),)
    hex_string_params = (("0x7fc00000", np.nan), ("0x7fc00001", np.nan), ("0x3f800000", 1.0))
    item_size_params = (Float32(),)


class TestFloat64(_BaseTestFloat):
    test_cls = Float64
    valid_dtype = (np.dtype(">f8"), np.dtype("<f8"))
    invalid_dtype = (
        np.dtype(np.int8),
        np.dtype(np.uint16),
        np.dtype(np.float32),
    )
    valid_json_v2 = (
        {"name": ">f8", "object_codec_id": None},
        {"name": "<f8", "object_codec_id": None},
    )
    valid_json_v3 = ("float64",)
    invalid_json_v2 = (
        "|f8",
        "float64",
        "|i1",
    )
    invalid_json_v3 = (
        "|f8",
        "|i1",
        {"name": "float64", "configuration": {"endianness": "little"}},
    )

    scalar_v2_params = (
        (Float64(), 1.0),
        (Float64(), -1.0),
        (Float64(), "NaN"),
        (Float64(), "Infinity"),
    )
    scalar_v3_params = (
        (Float64(), 1.0),
        (Float64(), -1.0),
        (Float64(), "NaN"),
        (Float64(), "Infinity"),
    )

    cast_value_params = (
        (Float64(), 1.0, np.float64(1.0)),
        (Float64(), -1.0, np.float64(-1.0)),
        (Float64(), "NaN", np.float64("NaN")),
    )
    invalid_scalar_params = ((Float64(), {"set!"}),)
    hex_string_params = (
        ("0x7ff8000000000000", np.nan),
        ("0x7ff8000000000001", np.nan),
        ("0x3ff0000000000000", 1.0),
    )
    item_size_params = (Float64(),)



import subprocess
from dataclasses import dataclass
from itertools import product
from pathlib import Path
from typing import TYPE_CHECKING, Literal

import numcodecs
import numpy as np
import pytest
from numcodecs import LZ4, LZMA, Blosc, GZip, VLenBytes, VLenUTF8, Zstd

import zarr
import zarr.abc
import zarr.abc.codec
import zarr.codecs as zarrcodecs
from zarr.core.array import Array
from zarr.core.chunk_key_encodings import V2ChunkKeyEncoding
from zarr.core.dtype.npy.bytes import VariableLengthBytes
from zarr.core.dtype.npy.string import VariableLengthUTF8
from zarr.storage import LocalStore

if TYPE_CHECKING:
    from zarr.core.dtype import ZDTypeLike

ZarrPythonVersion = Literal["2.18", "3.0.8"]


def runner_installed() -> bool:
    """
    Check if a PEP-723 compliant python script runner is installed.
    """
    try:
        subprocess.check_output(["uv", "--version"])
        return True  # noqa: TRY300
    except FileNotFoundError:
        return False


@dataclass(kw_only=True)
class ArrayParams:
    values: np.ndarray[tuple[int], np.dtype[np.generic]]
    fill_value: np.generic | str | int | bytes
    filters: tuple[numcodecs.abc.Codec, ...] = ()
    serializer: str | None = None
    compressor: numcodecs.abc.Codec


basic_codecs = GZip(), Blosc(), LZ4(), LZMA(), Zstd()
basic_dtypes = "|b", ">i2", ">i4", ">f4", ">f8", "<f4", "<f8", ">c8", "<c8", ">c16", "<c16"
datetime_dtypes = "<M8[10ns]", ">M8[10us]", "<m8[2ms]", ">m8[4ps]"
string_dtypes = "<U1", ">U4"
bytes_dtypes = ">S1", "<S4", ">V10", "<V4"

basic_array_cases = [
    ArrayParams(values=np.arange(4, dtype=dtype), fill_value=1, compressor=codec)
    for codec, dtype in product(basic_codecs, basic_dtypes)
]
datetime_array_cases = [
    ArrayParams(values=np.ones((4,), dtype=dtype), fill_value=1, compressor=codec)
    for codec, dtype in product(basic_codecs, datetime_dtypes)
]
string_array_cases = [
    ArrayParams(
        values=np.array(["aaaa", "bbbb", "ccccc", "dddd"], dtype=dtype),
        fill_value="foo",
        compressor=codec,
    )
    for codec, dtype in product(basic_codecs, string_dtypes)
]

bytes_array_cases = [
    ArrayParams(
        values=np.array([b"aaaa", b"bbbb", b"ccccc", b"dddd"], dtype=dtype),
        fill_value=b"foo",
        compressor=codec,
    )
    for codec, dtype in product(basic_codecs, bytes_dtypes)
]

vlen_string_cases = [
    ArrayParams(
        values=np.array(["a", "bb", "ccc", "dddd"], dtype="O"),
        fill_value="1",
        serializer="vlen-utf8",
        compressor=GZip(),
    )
]
vlen_bytes_cases = [
    ArrayParams(
        values=np.array([b"a", b"bb", b"ccc", b"dddd"], dtype="O"),
        fill_value=b"1",
        serializer="vlen-bytes",
        compressor=GZip(),
    )
]
array_cases_v2_18 = (
    basic_array_cases
    + bytes_array_cases
    + datetime_array_cases
    + string_array_cases
    + vlen_string_cases
    + vlen_bytes_cases
)

array_cases_v3_08 = vlen_string_cases


@pytest.fixture
def source_array_v2(tmp_path: Path, request: pytest.FixtureRequest) -> Array:
    """
    Writes a zarr array to a temporary directory based on the provided ArrayParams. The array is
    returned.
    """
    dest = tmp_path / "in"
    store = LocalStore(dest)
    array_params: ArrayParams = request.param
    compressor = array_params.compressor
    chunk_key_encoding = V2ChunkKeyEncoding(separator="/")
    dtype: ZDTypeLike
    if array_params.values.dtype == np.dtype("|O") and array_params.serializer == "vlen-utf8":
        dtype = VariableLengthUTF8()  # type: ignore[assignment]
        filters = array_params.filters + (VLenUTF8(),)
    elif array_params.values.dtype == np.dtype("|O") and array_params.serializer == "vlen-bytes":
        dtype = VariableLengthBytes()
        filters = array_params.filters + (VLenBytes(),)
    else:
        dtype = array_params.values.dtype
        filters = array_params.filters
    z = zarr.create_array(
        store,
        shape=array_params.values.shape,
        dtype=dtype,
        chunks=array_params.values.shape,
        compressors=compressor,
        filters=filters,
        fill_value=array_params.fill_value,
        order="C",
        chunk_key_encoding=chunk_key_encoding,
        write_data=True,
        zarr_format=2,
    )
    z[:] = array_params.values
    return z


@pytest.fixture
def source_array_v3(tmp_path: Path, request: pytest.FixtureRequest) -> Array:
    """
    Writes a zarr array to a temporary directory based on the provided ArrayParams. The array is
    returned.
    """
    dest = tmp_path / "in"
    store = LocalStore(dest)
    array_params: ArrayParams = request.param
    chunk_key_encoding = V2ChunkKeyEncoding(separator="/")
    dtype: ZDTypeLike
    serializer: Literal["auto"] | zarr.abc.codec.Codec
    if array_params.values.dtype == np.dtype("|O") and array_params.serializer == "vlen-utf8":
        dtype = VariableLengthUTF8()  # type: ignore[assignment]
        serializer = zarrcodecs.VLenUTF8Codec()
    elif array_params.values.dtype == np.dtype("|O") and array_params.serializer == "vlen-bytes":
        dtype = VariableLengthBytes()
        serializer = zarrcodecs.VLenBytesCodec()
    else:
        dtype = array_params.values.dtype
        serializer = "auto"
    if array_params.compressor == GZip():
        compressor = zarrcodecs.GzipCodec()
    else:
        msg = (
            "This test is only compatible with gzip compression at the moment, because the author"
            "did not want to implement a complete abstraction layer for v2 and v3 codecs in this test."
        )
        raise ValueError(msg)
    z = zarr.create_array(
        store,
        shape=array_params.values.shape,
        dtype=dtype,
        chunks=array_params.values.shape,
        compressors=compressor,
        filters=array_params.filters,
        serializer=serializer,
        fill_value=array_params.fill_value,
        chunk_key_encoding=chunk_key_encoding,
        write_data=True,
        zarr_format=3,
    )
    z[:] = array_params.values
    return z


# TODO: make this dynamic based on the installed scripts
script_paths = [Path(__file__).resolve().parent / "scripts" / "v2.18.py"]


@pytest.mark.skipif(not runner_installed(), reason="no python script runner installed")
@pytest.mark.parametrize(
    "source_array_v2", array_cases_v2_18, indirect=True, ids=tuple(map(str, array_cases_v2_18))
)
@pytest.mark.parametrize("script_path", script_paths)
def test_roundtrip_v2(source_array_v2: Array, tmp_path: Path, script_path: Path) -> None:
    out_path = tmp_path / "out"
    copy_op = subprocess.run(
        [
            "uv",
            "run",
            str(script_path),
            str(source_array_v2.store).removeprefix("file://"),
            str(out_path),
        ],
        capture_output=True,
        text=True,
    )
    assert copy_op.returncode == 0
    out_array = zarr.open_array(store=out_path, mode="r", zarr_format=2)
    assert source_array_v2.metadata.to_dict() == out_array.metadata.to_dict()
    assert np.array_equal(source_array_v2[:], out_array[:])


@pytest.mark.skipif(not runner_installed(), reason="no python script runner installed")
@pytest.mark.filterwarnings("ignore::zarr.core.dtype.common.UnstableSpecificationWarning")
@pytest.mark.parametrize(
    "source_array_v3", array_cases_v3_08, indirect=True, ids=tuple(map(str, array_cases_v3_08))
)
def test_roundtrip_v3(source_array_v3: Array, tmp_path: Path) -> None:
    script_path = Path(__file__).resolve().parent / "scripts" / "v3.0.8.py"
    out_path = tmp_path / "out"
    copy_op = subprocess.run(
        [
            "uv",
            "run",
            str(script_path),
            str(source_array_v3.store).removeprefix("file://"),
            str(out_path),
        ],
        capture_output=True,
        text=True,
    )
    assert copy_op.returncode == 0
    out_array = zarr.open_array(store=out_path, mode="r", zarr_format=3)
    assert source_array_v3.metadata.to_dict() == out_array.metadata.to_dict()
    assert np.array_equal(source_array_v3[:], out_array[:])



from __future__ import annotations

import warnings
from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from typing import (
    ClassVar,
    Final,
    Generic,
    Literal,
    TypedDict,
    TypeGuard,
    TypeVar,
)

from typing_extensions import ReadOnly

from zarr.core.common import NamedConfig

EndiannessStr = Literal["little", "big"]
ENDIANNESS_STR: Final = "little", "big"

SpecialFloatStrings = Literal["NaN", "Infinity", "-Infinity"]
SPECIAL_FLOAT_STRINGS: Final = ("NaN", "Infinity", "-Infinity")

JSONFloatV2 = float | SpecialFloatStrings
JSONFloatV3 = float | SpecialFloatStrings | str

ObjectCodecID = Literal["vlen-utf8", "vlen-bytes", "vlen-array", "pickle", "json2", "msgpack2"]
# These are the ids of the known object codecs for zarr v2.
OBJECT_CODEC_IDS: Final = ("vlen-utf8", "vlen-bytes", "vlen-array", "pickle", "json2", "msgpack2")

# This is a wider type than our standard JSON type because we need
# to work with typeddict objects which are assignable to Mapping[str, object]
DTypeJSON = str | int | float | Sequence["DTypeJSON"] | None | Mapping[str, object]

# The DTypeJSON_V2 type exists because ZDType.from_json takes a single argument, which must contain
# all the information necessary to decode the data type. Zarr v2 supports multiple distinct
# data types that all used the "|O" data type identifier. These data types can only be
# discriminated on the basis of their "object codec", i.e. a special data type specific
# compressor or filter. So to figure out what data type a zarr v2 array has, we need the
# data type identifier from metadata, as well as an object codec id if the data type identifier
# is "|O".
# So we will pack the name of the dtype alongside the name of the object codec id, if applicable,
# in a single dict, and pass that to the data type inference logic.
# These type variables have a very wide bound because the individual zdtype
# classes can perform a very specific type check.

# This is the JSON representation of a structured dtype in zarr v2
StructuredName_V2 = Sequence["str | StructuredName_V2"]

# This models the type of the name a dtype might have in zarr v2 array metadata
DTypeName_V2 = StructuredName_V2 | str

TDTypeNameV2_co = TypeVar("TDTypeNameV2_co", bound=DTypeName_V2, covariant=True)
TObjectCodecID_co = TypeVar("TObjectCodecID_co", bound=None | str, covariant=True)


class DTypeConfig_V2(TypedDict, Generic[TDTypeNameV2_co, TObjectCodecID_co]):
    name: ReadOnly[TDTypeNameV2_co]
    object_codec_id: ReadOnly[TObjectCodecID_co]


DTypeSpec_V2 = DTypeConfig_V2[DTypeName_V2, None | str]


def check_structured_dtype_v2_inner(data: object) -> TypeGuard[StructuredName_V2]:
    """
    A type guard for the inner elements of a structured dtype. This is a recursive check because
    the type is itself recursive.

    This check ensures that all the elements are 2-element sequences beginning with a string
    and ending with either another string or another 2-element sequence beginning with a string and
    ending with another instance of that type.
    """
    if isinstance(data, (str, Mapping)):
        return False
    if not isinstance(data, Sequence):
        return False
    if len(data) != 2:
        return False
    if not (isinstance(data[0], str)):
        return False
    if isinstance(data[-1], str):
        return True
    elif isinstance(data[-1], Sequence):
        return check_structured_dtype_v2_inner(data[-1])
    return False


def check_structured_dtype_name_v2(data: Sequence[object]) -> TypeGuard[StructuredName_V2]:
    """
    Check that all the elements of a sequence are valid zarr v2 structured dtype identifiers
    """
    return all(check_structured_dtype_v2_inner(d) for d in data)


def check_dtype_name_v2(data: object) -> TypeGuard[DTypeName_V2]:
    """
    Type guard for narrowing the type of a python object to an valid zarr v2 dtype name.
    """
    if isinstance(data, str):
        return True
    elif isinstance(data, Sequence):
        return check_structured_dtype_name_v2(data)
    return False


def check_dtype_spec_v2(data: object) -> TypeGuard[DTypeSpec_V2]:
    """
    Type guard for narrowing a python object to an instance of DTypeSpec_V2
    """
    if not isinstance(data, Mapping):
        return False
    if set(data.keys()) != {"name", "object_codec_id"}:
        return False
    if not check_dtype_name_v2(data["name"]):
        return False
    return isinstance(data["object_codec_id"], str | None)


# By comparison, The JSON representation of a dtype in zarr v3 is much simpler.
# It's either a string, or a structured dict
DTypeSpec_V3 = str | NamedConfig[str, Mapping[str, object]]


def check_dtype_spec_v3(data: object) -> TypeGuard[DTypeSpec_V3]:
    """
    Type guard for narrowing the type of a python object to an instance of
    DTypeSpec_V3, i.e either a string or a dict with a "name" field that's a string and a
    "configuration" field that's a mapping with string keys.
    """
    if isinstance(data, str) or (  # noqa: SIM103
        isinstance(data, Mapping)
        and set(data.keys()) == {"name", "configuration"}
        and isinstance(data["configuration"], Mapping)
        and all(isinstance(k, str) for k in data["configuration"])
    ):
        return True
    return False


def unpack_dtype_json(data: DTypeSpec_V2 | DTypeSpec_V3) -> DTypeJSON:
    """
    Return the array metadata form of the dtype JSON representation. For the Zarr V3 form of dtype
    metadata, this is a no-op. For the Zarr V2 form of dtype metadata, this unpacks the dtype name.
    """
    if isinstance(data, Mapping) and set(data.keys()) == {"name", "object_codec_id"}:
        return data["name"]
    return data


class DataTypeValidationError(ValueError): ...


class ScalarTypeValidationError(ValueError): ...


@dataclass(frozen=True, kw_only=True)
class HasLength:
    """
    A mix-in class for data types with a length attribute, such as fixed-size collections
    of unicode strings, or bytes.

    Attributes
    ----------
    length : int
        The length of the scalars belonging to this data type. Note that this class does not assign
        a unit to the length. Child classes may assign units.
    """

    length: int


@dataclass(frozen=True, kw_only=True)
class HasEndianness:
    """
    A mix-in class for data types with an endianness attribute
    """

    endianness: EndiannessStr = "little"


@dataclass(frozen=True, kw_only=True)
class HasItemSize:
    """
    A mix-in class for data types with an item size attribute.
    This mix-in bears a property ``item_size``, which denotes the size of each element of the data
    type, in bytes.
    """

    @property
    def item_size(self) -> int:
        raise NotImplementedError


@dataclass(frozen=True, kw_only=True)
class HasObjectCodec:
    """
    A mix-in class for data types that require an object codec id.
    This class bears the property ``object_codec_id``, which is the string name of an object
    codec that is required to encode and decode the data type.

    In zarr-python 2.x certain data types like variable-length strings or variable-length arrays
    used the catch-all numpy "object" data type for their in-memory representation. But these data
    types cannot be stored as numpy object data types, because the object data type does not define
    a fixed memory layout. So these data types required a special codec, called an "object codec",
    that effectively defined a compact representation for the data type, which was used to encode
    and decode the data type.

    Zarr-python 2.x would not allow the creation of arrays with the "object" data type if an object
    codec was not specified, and thus the name of the object codec is effectively part of the data
    type model.
    """

    object_codec_id: ClassVar[str]


class UnstableSpecificationWarning(FutureWarning): ...


def v3_unstable_dtype_warning(dtype: object) -> None:
    """
    Emit this warning when a data type does not have a stable zarr v3 spec
    """
    msg = (
        f"The data type ({dtype}) does not have a Zarr V3 specification. "
        "That means that the representation of arrays saved with this data type may change without "
        "warning in a future version of Zarr Python. "
        "Arrays stored with this data type may be unreadable by other Zarr libraries. "
        "Use this data type at your own risk! "
        "Check https://github.com/zarr-developers/zarr-extensions/tree/main/data-types for the "
        "status of data type specifications for Zarr V3."
    )
    warnings.warn(msg, category=UnstableSpecificationWarning, stacklevel=2)



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import numpy.typing as npt

import zarr.core.buffer
from zarr.abc.codec import ArrayBytesCodec, CodecInput, CodecPipeline
from zarr.codecs import BytesCodec
from zarr.core.buffer import Buffer, NDBuffer
from zarr.core.dtype.common import DataTypeValidationError, DTypeJSON, DTypeSpec_V2
from zarr.core.dtype.npy.bool import Bool

if TYPE_CHECKING:
    from collections.abc import Iterable
    from typing import Any, ClassVar, Literal, Self

    from zarr.core.array_spec import ArraySpec
    from zarr.core.common import ZarrFormat


class TestEntrypointCodec(ArrayBytesCodec):
    is_fixed_size = True

    async def encode(
        self,
        chunks_and_specs: Iterable[tuple[CodecInput | None, ArraySpec]],
    ) -> Iterable[Buffer | None]:
        return [None]

    async def decode(
        self,
        chunks_and_specs: Iterable[tuple[CodecInput | None, ArraySpec]],
    ) -> npt.NDArray[Any]:
        return np.array(1)

    def compute_encoded_size(self, input_byte_length: int, chunk_spec: ArraySpec) -> int:
        return input_byte_length


class TestEntrypointCodecPipeline(CodecPipeline):
    def __init__(self, batch_size: int = 1) -> None:
        pass

    async def encode(
        self, chunks_and_specs: Iterable[tuple[CodecInput | None, ArraySpec]]
    ) -> Iterable[Buffer | None]:
        return [None]

    async def decode(
        self, chunks_and_specs: Iterable[tuple[CodecInput | None, ArraySpec]]
    ) -> Iterable[NDBuffer | None]:
        return np.array(1)


class TestEntrypointBuffer(Buffer):
    pass


class TestEntrypointNDBuffer(NDBuffer):
    pass


class TestEntrypointGroup:
    class Codec(BytesCodec):
        pass

    class Buffer(zarr.core.buffer.Buffer):
        pass

    class NDBuffer(zarr.core.buffer.NDBuffer):
        pass

    class Pipeline(CodecPipeline):
        pass


class TestDataType(Bool):
    """
    This is a "data type" that serializes to "test"
    """

    _zarr_v3_name: ClassVar[Literal["test"]] = "test"  # type: ignore[assignment]

    @classmethod
    def from_json(cls, data: DTypeJSON, *, zarr_format: Literal[2, 3]) -> Self:
        if zarr_format == 2 and data == {"name": cls._zarr_v3_name, "object_codec_id": None}:
            return cls()
        if zarr_format == 3 and data == cls._zarr_v3_name:
            return cls()
        raise DataTypeValidationError(
            f"Invalid JSON representation of {cls.__name__}. Got {data!r}"
        )

    def to_json(self, zarr_format: ZarrFormat) -> str | DTypeSpec_V2:  # type: ignore[override]
        if zarr_format == 2:
            return {"name": self._zarr_v3_name, "object_codec_id": None}
        if zarr_format == 3:
            return self._zarr_v3_name
        raise ValueError("zarr_format must be 2 or 3")



"""
Wrapper for native array data types.

The ``ZDType`` class is an abstract base class for wrapping native array data types, e.g. NumPy dtypes.
``ZDType`` provides a common interface for working with data types in a way that is independent of the
underlying data type system.

The wrapper class encapsulates a native data type. Instances of the class can be created from a
native data type instance, and a native data type instance can be created from an instance of the
wrapper class.

The wrapper class is responsible for:
- Serializing and deserializing a native data type to Zarr V2 or Zarr V3 metadata.
  This ensures that the data type can be properly stored and retrieved from array metadata.
- Serializing and deserializing scalar values to Zarr V2 or Zarr V3 metadata. This is important for
  storing a fill value for an array in a manner that is valid for the data type.

You can add support for a new data type in Zarr by subclassing ``ZDType`` wrapper class and adapt its methods
to support your native data type. The wrapper class must be added to a data type registry
(defined elsewhere) before array creation routines or array reading routines can use your new data
type.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    ClassVar,
    Generic,
    Literal,
    Self,
    TypeGuard,
    TypeVar,
    overload,
)

import numpy as np

if TYPE_CHECKING:
    from zarr.core.common import JSON, ZarrFormat
    from zarr.core.dtype.common import DTypeJSON, DTypeSpec_V2, DTypeSpec_V3

# This the upper bound for the scalar types we support. It's numpy scalars + str,
# because the new variable-length string dtype in numpy does not have a corresponding scalar type
TBaseScalar = np.generic | str | bytes
# This is the bound for the dtypes that we support. If we support non-numpy dtypes,
# then this bound will need to be widened.
TBaseDType = np.dtype[np.generic]

# These two type parameters are covariant because we want
# x : ZDType[BaseDType, BaseScalar] = ZDType[SubDType, SubScalar]
# to type check
TScalar_co = TypeVar("TScalar_co", bound=TBaseScalar, covariant=True)
TDType_co = TypeVar("TDType_co", bound=TBaseDType, covariant=True)


@dataclass(frozen=True, kw_only=True, slots=True)
class ZDType(ABC, Generic[TDType_co, TScalar_co]):
    """
    Abstract base class for wrapping native array data types, e.g. numpy dtypes

    Attributes
    ----------
    dtype_cls : ClassVar[type[TDType]]
        The wrapped dtype class. This is a class variable.
    _zarr_v3_name : ClassVar[str]
        The name given to the data type by a Zarr v3 data type specification. This is a
        class variable, and it should generally be unique across different data types.
    """

    # this class will create a native data type
    # mypy currently disallows class variables to contain type parameters
    # but it seems OK for us to use it here:
    # https://github.com/python/typing/discussions/1424#discussioncomment-7989934
    dtype_cls: ClassVar[type[TDType_co]]  # type: ignore[misc]
    _zarr_v3_name: ClassVar[str]

    @classmethod
    def _check_native_dtype(cls: type[Self], dtype: TBaseDType) -> TypeGuard[TDType_co]:
        """
        Check that a native data type matches the dtype_cls class attribute.

        Used as a type guard.

        Parameters
        ----------
        dtype : TDType
            The dtype to check.

        Returns
        -------
        Bool
            True if the dtype matches, False otherwise.
        """
        return type(dtype) is cls.dtype_cls

    @classmethod
    @abstractmethod
    def from_native_dtype(cls: type[Self], dtype: TBaseDType) -> Self:
        """
        Create a ZDType instance from a native data type.

        This method is used when taking a user-provided native data type, like a NumPy data type,
        and creating the corresponding ZDType instance from them.

        Parameters
        ----------
        dtype : TDType
            The native data type object to wrap.

        Returns
        -------
        Self
            The ZDType that wraps the native data type.

        Raises
        ------
        TypeError
            If the native data type is not consistent with the wrapped data type.
        """
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def to_native_dtype(self: Self) -> TDType_co:
        """
        Return an instance of the wrapped data type. This operation inverts ``from_native_dtype``.

        Returns
        -------
        TDType
            The native data type wrapped by this ZDType.
        """
        raise NotImplementedError  # pragma: no cover

    @classmethod
    @abstractmethod
    def _from_json_v2(cls: type[Self], data: DTypeJSON) -> Self:
        raise NotImplementedError  # pragma: no cover

    @classmethod
    @abstractmethod
    def _from_json_v3(cls: type[Self], data: DTypeJSON) -> Self:
        raise NotImplementedError  # pragma: no cover

    @classmethod
    def from_json(cls: type[Self], data: DTypeJSON, *, zarr_format: ZarrFormat) -> Self:
        """
        Create an instance of this ZDType from JSON data.

        Parameters
        ----------
        data : DTypeJSON
            The JSON representation of the data type.

        zarr_format : ZarrFormat
            The zarr format version.

        Returns
        -------
        Self
            An instance of this data type.
        """
        if zarr_format == 2:
            return cls._from_json_v2(data)
        if zarr_format == 3:
            return cls._from_json_v3(data)
        raise ValueError(f"zarr_format must be 2 or 3, got {zarr_format}")  # pragma: no cover

    @overload
    def to_json(self, zarr_format: Literal[2]) -> DTypeSpec_V2: ...

    @overload
    def to_json(self, zarr_format: Literal[3]) -> DTypeSpec_V3: ...

    @abstractmethod
    def to_json(self, zarr_format: ZarrFormat) -> DTypeSpec_V2 | DTypeSpec_V3:
        """
        Serialize this ZDType to JSON.

        Parameters
        ----------
        zarr_format : ZarrFormat
            The zarr format version.

        Returns
        -------
        DTypeJSON_V2 | DTypeJSON_V3
            The JSON-serializable representation of the wrapped data type
        """
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def _check_scalar(self, data: object) -> bool:
        """
        Check that an python object is a valid scalar value for the wrapped data type.

        Parameters
        ----------
        data : object
            A value to check.

        Returns
        -------
        Bool
            True if the object is valid, False otherwise.
        """
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def cast_scalar(self, data: object) -> TScalar_co:
        """
        Cast a python object to the wrapped scalar type.

        The type of the provided scalar is first checked for compatibility.
        If it's incompatible with the associated scalar type, a ``TypeError`` will be raised.

        Parameters
        ----------
        data : object
            The python object to cast.

        Returns
        -------
        TScalar
            The cast value.
        """
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def default_scalar(self) -> TScalar_co:
        """
        Get the default scalar value for the wrapped data type.

        This is a method, rather than an attribute, because the default value for some data types depends on parameters that are
        not known until a concrete data type is wrapped. For example, data types parametrized by a
        length like fixed-length strings or bytes will generate scalars consistent with that length.

        Returns
        -------
        TScalar
            The default value for this data type.
        """
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def from_json_scalar(self: Self, data: JSON, *, zarr_format: ZarrFormat) -> TScalar_co:
        """
        Read a JSON-serializable value as a scalar.

        Parameters
        ----------
        data : JSON
            A JSON representation of a scalar value.
        zarr_format : ZarrFormat
            The zarr format version. This is specified because the JSON serialization of scalars
            differs between Zarr V2 and Zarr V3.

        Returns
        -------
        TScalar
            The deserialized scalar value.
        """
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def to_json_scalar(self, data: object, *, zarr_format: ZarrFormat) -> JSON:
        """
        Serialize a python object to the JSON representation of a scalar.

        The value will first be cast to the scalar type associated with this ZDType, then serialized
        to JSON.

        Parameters
        ----------
        data : object
            The value to convert.
        zarr_format : ZarrFormat
            The zarr format version. This is specified because the JSON serialization of scalars
            differs between Zarr V2 and Zarr V3.

        Returns
        -------
        JSON
            The JSON-serialized scalar.
        """
        raise NotImplementedError  # pragma: no cover


def scalar_failed_type_check_msg(
    cls_instance: ZDType[TBaseDType, TBaseScalar], bad_scalar: object
) -> str:
    """
    Generate an error message reporting that a particular value failed a type check when attempting
    to cast that value to a scalar.
    """
    return (
        f"The value {bad_scalar!r} failed a type check. "
        f"It cannot be safely cast to a scalar compatible with {cls_instance}. "
        f"Consult the documentation for {cls_instance} to determine the possible values that can "
        "be cast to scalars of the wrapped data type."
    )



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, ClassVar, Literal, Self, TypeGuard, overload

import numpy as np

from zarr.core.dtype.common import (
    DataTypeValidationError,
    DTypeConfig_V2,
    DTypeJSON,
    HasEndianness,
    HasItemSize,
    check_dtype_spec_v2,
)
from zarr.core.dtype.npy.common import (
    FloatLike,
    TFloatDType_co,
    TFloatScalar_co,
    check_json_float_v2,
    check_json_float_v3,
    endianness_to_numpy_str,
    float_from_json_v2,
    float_from_json_v3,
    float_to_json_v2,
    float_to_json_v3,
    get_endianness_from_numpy_dtype,
)
from zarr.core.dtype.wrapper import TBaseDType, ZDType

if TYPE_CHECKING:
    from zarr.core.common import JSON, ZarrFormat


@dataclass(frozen=True)
class BaseFloat(ZDType[TFloatDType_co, TFloatScalar_co], HasEndianness, HasItemSize):
    """
    A base class for Zarr data types that wrap NumPy float data types.
    """

    # This attribute holds the possible zarr v2 JSON names for the data type
    _zarr_v2_names: ClassVar[tuple[str, ...]]

    @classmethod
    def from_native_dtype(cls, dtype: TBaseDType) -> Self:
        """
        Create an instance of this ZDType from a NumPy data type.

        Parameters
        ----------
        dtype : TBaseDType
            The NumPy data type.

        Returns
        -------
        Self
            An instance of this data type.
        """
        if cls._check_native_dtype(dtype):
            return cls(endianness=get_endianness_from_numpy_dtype(dtype))
        raise DataTypeValidationError(
            f"Invalid data type: {dtype}. Expected an instance of {cls.dtype_cls}"
        )

    def to_native_dtype(self) -> TFloatDType_co:
        """
        Convert the wrapped data type to a NumPy data type.

        Returns
        -------
        TFloatDType_co
            The NumPy data type.
        """
        byte_order = endianness_to_numpy_str(self.endianness)
        return self.dtype_cls().newbyteorder(byte_order)  # type: ignore[return-value]

    @classmethod
    def _check_json_v2(cls, data: DTypeJSON) -> TypeGuard[DTypeConfig_V2[str, None]]:
        """
        Check that the input is a valid JSON representation of this data type.

        Parameters
        ----------
        data : DTypeJSON
            The JSON data.

        Returns
        -------
        TypeGuard[DTypeConfig_V2[str, None]]
            True if the input is a valid JSON representation of this data type, False otherwise.
        """
        return (
            check_dtype_spec_v2(data)
            and data["name"] in cls._zarr_v2_names
            and data["object_codec_id"] is None
        )

    @classmethod
    def _check_json_v3(cls, data: DTypeJSON) -> TypeGuard[str]:
        """
        Check that the input is a valid JSON representation of this class in Zarr V3.

        Parameters
        ----------
        data : DTypeJSON
            The JSON data.

        Returns
        -------
        TypeGuard[str]
            True if the input is a valid JSON representation of this class, False otherwise.
        """
        return data == cls._zarr_v3_name

    @classmethod
    def _from_json_v2(cls, data: DTypeJSON) -> Self:
        """
        Create an instance of this ZDType from Zarr v2-flavored JSON.

        Parameters
        ----------
        data : DTypeJSON
            The JSON data.

        Returns
        -------
        Self
            An instance of this data type.
        """
        if cls._check_json_v2(data):
            # Going via NumPy ensures that we get the endianness correct without
            # annoying string parsing.
            name = data["name"]
            return cls.from_native_dtype(np.dtype(name))
        msg = f"Invalid JSON representation of {cls.__name__}. Got {data!r}, expected one of the strings {cls._zarr_v2_names}."
        raise DataTypeValidationError(msg)

    @classmethod
    def _from_json_v3(cls, data: DTypeJSON) -> Self:
        """
        Create an instance of this ZDType from Zarr v3-flavored JSON.

        Parameters
        ----------
        data : DTypeJSON
            The JSON data.

        Returns
        -------
        Self
            An instance of this data type.
        """
        if cls._check_json_v3(data):
            return cls()
        msg = f"Invalid JSON representation of {cls.__name__}. Got {data!r}, expected {cls._zarr_v3_name}."
        raise DataTypeValidationError(msg)

    @overload
    def to_json(self, zarr_format: Literal[2]) -> DTypeConfig_V2[str, None]: ...

    @overload
    def to_json(self, zarr_format: Literal[3]) -> str: ...

    def to_json(self, zarr_format: ZarrFormat) -> DTypeConfig_V2[str, None] | str:
        """
        Convert the wrapped data type to a JSON-serializable form.

        Parameters
        ----------
        zarr_format : ZarrFormat
            The zarr format version.

        Returns
        -------
        DTypeConfig_V2[str, None] or str
            The JSON-serializable representation of the wrapped data type.

        Raises
        ------
        ValueError
            If zarr_format is not 2 or 3.
        """
        if zarr_format == 2:
            return {"name": self.to_native_dtype().str, "object_codec_id": None}
        elif zarr_format == 3:
            return self._zarr_v3_name
        raise ValueError(f"zarr_format must be 2 or 3, got {zarr_format}")  # pragma: no cover

    def _check_scalar(self, data: object) -> TypeGuard[FloatLike]:
        """
        Check that the input is a valid scalar value.

        Parameters
        ----------
        data : object
            The input to check.

        Returns
        -------
        TypeGuard[FloatLike]
            True if the input is a valid scalar value, False otherwise.
        """
        return isinstance(data, FloatLike)

    def _cast_scalar_unchecked(self, data: FloatLike) -> TFloatScalar_co:
        """
        Cast a scalar value to a NumPy float scalar.

        Parameters
        ----------
        data : FloatLike
            The scalar value to cast.

        Returns
        -------
        TFloatScalar_co
            The NumPy float scalar.
        """
        return self.to_native_dtype().type(data)  # type: ignore[return-value]

    def cast_scalar(self, data: object) -> TFloatScalar_co:
        """
        Cast a scalar value to a NumPy float scalar.

        Parameters
        ----------
        data : object
            The scalar value to cast.

        Returns
        -------
        TFloatScalar_co
            The NumPy float scalar.
        """
        if self._check_scalar(data):
            return self._cast_scalar_unchecked(data)
        msg = (
            f"Cannot convert object {data!r} with type {type(data)} to a scalar compatible with the "
            f"data type {self}."
        )
        raise TypeError(msg)

    def default_scalar(self) -> TFloatScalar_co:
        """
        Get the default value, which is 0 cast to this zdtype.

        Returns
        -------
        TFloatScalar_co
            The default value.
        """
        return self._cast_scalar_unchecked(0)

    def from_json_scalar(self, data: JSON, *, zarr_format: ZarrFormat) -> TFloatScalar_co:
        """
        Read a JSON-serializable value as a NumPy float scalar.

        Parameters
        ----------
        data : JSON
            The JSON-serializable value.
        zarr_format : ZarrFormat
            The zarr format version.

        Returns
        -------
        TFloatScalar_co
            The NumPy float scalar.
        """
        if zarr_format == 2:
            if check_json_float_v2(data):
                return self._cast_scalar_unchecked(float_from_json_v2(data))
            else:
                raise TypeError(
                    f"Invalid type: {data}. Expected a float or a special string encoding of a float."
                )
        elif zarr_format == 3:
            if check_json_float_v3(data):
                return self._cast_scalar_unchecked(float_from_json_v3(data))
            else:
                raise TypeError(
                    f"Invalid type: {data}. Expected a float or a special string encoding of a float."
                )
        else:
            raise ValueError(f"zarr_format must be 2 or 3, got {zarr_format}")  # pragma: no cover

    def to_json_scalar(self, data: object, *, zarr_format: ZarrFormat) -> float | str:
        """
        Convert an object to a JSON-serializable float.

        Parameters
        ----------
        data : _BaseScalar
            The value to convert.
        zarr_format : ZarrFormat
            The zarr format version.

        Returns
        -------
        JSON
            The JSON-serializable form of the float, which is potentially a number or a string.
            See the zarr specifications for details on the JSON encoding for floats.
        """
        if zarr_format == 2:
            return float_to_json_v2(self.cast_scalar(data))
        elif zarr_format == 3:
            return float_to_json_v3(self.cast_scalar(data))
        else:
            raise ValueError(f"zarr_format must be 2 or 3, got {zarr_format}")  # pragma: no cover


@dataclass(frozen=True, kw_only=True)
class Float16(BaseFloat[np.dtypes.Float16DType, np.float16]):
    """
    A Zarr data type for arrays containing 16-bit floating point numbers.

    Wraps the ``np.dtypes.Float16DType`` data type. Scalars for this data type are instances
    of ``np.float16``.

    Attributes
    ----------
    dtype_cls : Type[np.dtypes.Float16DType]
        The NumPy dtype class for this data type.

    References
    ----------
    This class implements the float16 data type defined in Zarr V2 and V3.

    See the `Zarr V2 <https://github.com/zarr-developers/zarr-specs/blob/main/docs/v2/v2.0.rst#data-type-encoding>`__ and `Zarr V3 <https://github.com/zarr-developers/zarr-specs/blob/main/docs/v3/data-types/index.rst>`__ specification documents for details.
    """

    dtype_cls = np.dtypes.Float16DType
    _zarr_v3_name = "float16"
    _zarr_v2_names: ClassVar[tuple[Literal[">f2"], Literal["<f2"]]] = (">f2", "<f2")

    @property
    def item_size(self) -> int:
        """
        The size of a single scalar in bytes.

        Returns
        -------
        int
            The size of a single scalar in bytes.
        """
        return 2


@dataclass(frozen=True, kw_only=True)
class Float32(BaseFloat[np.dtypes.Float32DType, np.float32]):
    """
    A Zarr data type for arrays containing 32-bit floating point numbers.

    Wraps the ``np.dtypes.Float32DType`` data type. Scalars for this data type are instances
    of ``np.float32``.

    Attributes
    ----------
    dtype_cls : Type[np.dtypes.Float32DType]
        The NumPy dtype class for this data type.

    References
    ----------
    This class implements the float32 data type defined in Zarr V2 and V3.

    See the `Zarr V2 <https://github.com/zarr-developers/zarr-specs/blob/main/docs/v2/v2.0.rst#data-type-encoding>`__ and `Zarr V3 <https://github.com/zarr-developers/zarr-specs/blob/main/docs/v3/data-types/index.rst>`__ specification documents for details.
    """

    dtype_cls = np.dtypes.Float32DType
    _zarr_v3_name = "float32"
    _zarr_v2_names: ClassVar[tuple[Literal[">f4"], Literal["<f4"]]] = (">f4", "<f4")

    @property
    def item_size(self) -> int:
        """
        The size of a single scalar in bytes.

        Returns
        -------
        int
            The size of a single scalar in bytes.
        """
        return 4


@dataclass(frozen=True, kw_only=True)
class Float64(BaseFloat[np.dtypes.Float64DType, np.float64]):
    """
    A Zarr data type for arrays containing 64-bit floating point numbers.

    Wraps the ``np.dtypes.Float64DType`` data type. Scalars for this data type are instances
    of ``np.float64``.

    Attributes
    ----------
    dtype_cls : Type[np.dtypes.Float64DType]
        The NumPy dtype class for this data type.

    References
    ----------
    This class implements the float64 data type defined in Zarr V2 and V3.

    See the `Zarr V2 <https://github.com/zarr-developers/zarr-specs/blob/main/docs/v2/v2.0.rst#data-type-encoding>`__ and `Zarr V3 <https://github.com/zarr-developers/zarr-specs/blob/main/docs/v3/data-types/index.rst>`__ specification documents for details.
    """

    dtype_cls = np.dtypes.Float64DType
    _zarr_v3_name = "float64"
    _zarr_v2_names: ClassVar[tuple[Literal[">f8"], Literal["<f8"]]] = (">f8", "<f8")

    @property
    def item_size(self) -> int:
        """
        The size of a single scalar in bytes.

        Returns
        -------
        int
            The size of a single scalar in bytes.
        """
        return 8



from __future__ import annotations

import warnings
from collections import defaultdict
from importlib.metadata import entry_points as get_entry_points
from typing import TYPE_CHECKING, Any, Generic, TypeVar

from zarr.core.config import BadConfigError, config
from zarr.core.dtype import data_type_registry

if TYPE_CHECKING:
    from importlib.metadata import EntryPoint

    from zarr.abc.codec import (
        ArrayArrayCodec,
        ArrayBytesCodec,
        BytesBytesCodec,
        Codec,
        CodecPipeline,
    )
    from zarr.core.buffer import Buffer, NDBuffer
    from zarr.core.common import JSON

__all__ = [
    "Registry",
    "get_buffer_class",
    "get_codec_class",
    "get_ndbuffer_class",
    "get_pipeline_class",
    "register_buffer",
    "register_codec",
    "register_ndbuffer",
    "register_pipeline",
]

T = TypeVar("T")


class Registry(dict[str, type[T]], Generic[T]):
    def __init__(self) -> None:
        super().__init__()
        self.lazy_load_list: list[EntryPoint] = []

    def lazy_load(self) -> None:
        for e in self.lazy_load_list:
            self.register(e.load())

        self.lazy_load_list.clear()

    def register(self, cls: type[T], qualname: str | None = None) -> None:
        if qualname is None:
            qualname = fully_qualified_name(cls)
        self[qualname] = cls


__codec_registries: dict[str, Registry[Codec]] = defaultdict(Registry)
__pipeline_registry: Registry[CodecPipeline] = Registry()
__buffer_registry: Registry[Buffer] = Registry()
__ndbuffer_registry: Registry[NDBuffer] = Registry()

"""
The registry module is responsible for managing implementations of codecs,
pipelines, buffers and ndbuffers and collecting them from entrypoints.
The implementation used is determined by the config.

The registry module is also responsible for managing dtypes.
"""


def _collect_entrypoints() -> list[Registry[Any]]:
    """
    Collects codecs, pipelines, dtypes, buffers and ndbuffers from entrypoints.
    Entry points can either be single items or groups of items.
    Allowed syntax for entry_points.txt is e.g.

        [zarr.codecs]
        gzip = package:EntrypointGzipCodec1
        [zarr.codecs.gzip]
        some_name = package:EntrypointGzipCodec2
        another = package:EntrypointGzipCodec3

        [zarr]
        buffer = package:TestBuffer1
        [zarr.buffer]
        xyz = package:TestBuffer2
        abc = package:TestBuffer3
        ...
    """
    entry_points = get_entry_points()

    __buffer_registry.lazy_load_list.extend(entry_points.select(group="zarr.buffer"))
    __buffer_registry.lazy_load_list.extend(entry_points.select(group="zarr", name="buffer"))
    __ndbuffer_registry.lazy_load_list.extend(entry_points.select(group="zarr.ndbuffer"))
    __ndbuffer_registry.lazy_load_list.extend(entry_points.select(group="zarr", name="ndbuffer"))

    data_type_registry._lazy_load_list.extend(entry_points.select(group="zarr.data_type"))
    data_type_registry._lazy_load_list.extend(entry_points.select(group="zarr", name="data_type"))

    __pipeline_registry.lazy_load_list.extend(entry_points.select(group="zarr.codec_pipeline"))
    __pipeline_registry.lazy_load_list.extend(
        entry_points.select(group="zarr", name="codec_pipeline")
    )
    for e in entry_points.select(group="zarr.codecs"):
        __codec_registries[e.name].lazy_load_list.append(e)
    for group in entry_points.groups:
        if group.startswith("zarr.codecs."):
            codec_name = group.split(".")[2]
            __codec_registries[codec_name].lazy_load_list.extend(entry_points.select(group=group))
    return [
        *__codec_registries.values(),
        __pipeline_registry,
        __buffer_registry,
        __ndbuffer_registry,
    ]


def _reload_config() -> None:
    config.refresh()


def fully_qualified_name(cls: type) -> str:
    module = cls.__module__
    return module + "." + cls.__qualname__


def register_codec(key: str, codec_cls: type[Codec]) -> None:
    if key not in __codec_registries:
        __codec_registries[key] = Registry()
    __codec_registries[key].register(codec_cls)


def register_pipeline(pipe_cls: type[CodecPipeline]) -> None:
    __pipeline_registry.register(pipe_cls)


def register_ndbuffer(cls: type[NDBuffer], qualname: str | None = None) -> None:
    __ndbuffer_registry.register(cls, qualname)


def register_buffer(cls: type[Buffer], qualname: str | None = None) -> None:
    __buffer_registry.register(cls, qualname)


def get_codec_class(key: str, reload_config: bool = False) -> type[Codec]:
    if reload_config:
        _reload_config()

    if key in __codec_registries:
        # logger.debug("Auto loading codec '%s' from entrypoint", codec_id)
        __codec_registries[key].lazy_load()

    codec_classes = __codec_registries[key]
    if not codec_classes:
        raise KeyError(key)

    config_entry = config.get("codecs", {}).get(key)
    if config_entry is None:
        if len(codec_classes) == 1:
            return next(iter(codec_classes.values()))
        warnings.warn(
            f"Codec '{key}' not configured in config. Selecting any implementation.",
            stacklevel=2,
        )
        return list(codec_classes.values())[-1]
    selected_codec_cls = codec_classes[config_entry]

    if selected_codec_cls:
        return selected_codec_cls
    raise KeyError(key)


def _resolve_codec(data: dict[str, JSON]) -> Codec:
    """
    Get a codec instance from a dict representation of that codec.
    """
    # TODO: narrow the type of the input to only those dicts that map on to codec class instances.
    return get_codec_class(data["name"]).from_dict(data)  # type: ignore[arg-type]


def _parse_bytes_bytes_codec(data: dict[str, JSON] | Codec) -> BytesBytesCodec:
    """
    Normalize the input to a ``BytesBytesCodec`` instance.
    If the input is already a ``BytesBytesCodec``, it is returned as is. If the input is a dict, it
    is converted to a ``BytesBytesCodec`` instance via the ``_resolve_codec`` function.
    """
    from zarr.abc.codec import BytesBytesCodec

    if isinstance(data, dict):
        result = _resolve_codec(data)
        if not isinstance(result, BytesBytesCodec):
            msg = f"Expected a dict representation of a BytesBytesCodec; got a dict representation of a {type(result)} instead."
            raise TypeError(msg)
    else:
        if not isinstance(data, BytesBytesCodec):
            raise TypeError(f"Expected a BytesBytesCodec. Got {type(data)} instead.")
        result = data
    return result


def _parse_array_bytes_codec(data: dict[str, JSON] | Codec) -> ArrayBytesCodec:
    """
    Normalize the input to a ``ArrayBytesCodec`` instance.
    If the input is already a ``ArrayBytesCodec``, it is returned as is. If the input is a dict, it
    is converted to a ``ArrayBytesCodec`` instance via the ``_resolve_codec`` function.
    """
    from zarr.abc.codec import ArrayBytesCodec

    if isinstance(data, dict):
        result = _resolve_codec(data)
        if not isinstance(result, ArrayBytesCodec):
            msg = f"Expected a dict representation of a ArrayBytesCodec; got a dict representation of a {type(result)} instead."
            raise TypeError(msg)
    else:
        if not isinstance(data, ArrayBytesCodec):
            raise TypeError(f"Expected a ArrayBytesCodec. Got {type(data)} instead.")
        result = data
    return result


def _parse_array_array_codec(data: dict[str, JSON] | Codec) -> ArrayArrayCodec:
    """
    Normalize the input to a ``ArrayArrayCodec`` instance.
    If the input is already a ``ArrayArrayCodec``, it is returned as is. If the input is a dict, it
    is converted to a ``ArrayArrayCodec`` instance via the ``_resolve_codec`` function.
    """
    from zarr.abc.codec import ArrayArrayCodec

    if isinstance(data, dict):
        result = _resolve_codec(data)
        if not isinstance(result, ArrayArrayCodec):
            msg = f"Expected a dict representation of a ArrayArrayCodec; got a dict representation of a {type(result)} instead."
            raise TypeError(msg)
    else:
        if not isinstance(data, ArrayArrayCodec):
            raise TypeError(f"Expected a ArrayArrayCodec. Got {type(data)} instead.")
        result = data
    return result


def get_pipeline_class(reload_config: bool = False) -> type[CodecPipeline]:
    if reload_config:
        _reload_config()
    __pipeline_registry.lazy_load()
    path = config.get("codec_pipeline.path")
    pipeline_class = __pipeline_registry.get(path)
    if pipeline_class:
        return pipeline_class
    raise BadConfigError(
        f"Pipeline class '{path}' not found in registered pipelines: {list(__pipeline_registry)}."
    )


def get_buffer_class(reload_config: bool = False) -> type[Buffer]:
    if reload_config:
        _reload_config()
    __buffer_registry.lazy_load()

    path = config.get("buffer")
    buffer_class = __buffer_registry.get(path)
    if buffer_class:
        return buffer_class
    raise BadConfigError(
        f"Buffer class '{path}' not found in registered buffers: {list(__buffer_registry)}."
    )


def get_ndbuffer_class(reload_config: bool = False) -> type[NDBuffer]:
    if reload_config:
        _reload_config()
    __ndbuffer_registry.lazy_load()
    path = config.get("ndbuffer")
    ndbuffer_class = __ndbuffer_registry.get(path)
    if ndbuffer_class:
        return ndbuffer_class
    raise BadConfigError(
        f"NDBuffer class '{path}' not found in registered buffers: {list(__ndbuffer_registry)}."
    )


_collect_entrypoints()



from typing import Any

import numpy as np
import pytest

import zarr
from zarr import Array
from zarr.abc.codec import Codec
from zarr.abc.store import Store
from zarr.codecs import ZstdCodec
from zarr.core.dtype import get_data_type_from_native_dtype
from zarr.core.dtype.npy.string import _NUMPY_SUPPORTS_VLEN_STRING
from zarr.core.metadata.v3 import ArrayV3Metadata
from zarr.storage import StorePath

numpy_str_dtypes: list[type | str | None] = [None, str, "str", np.dtypes.StrDType, "S", "U"]
expected_array_string_dtype: np.dtype[Any]
if _NUMPY_SUPPORTS_VLEN_STRING:
    numpy_str_dtypes.append(np.dtypes.StringDType)
    expected_array_string_dtype = np.dtypes.StringDType()
else:
    expected_array_string_dtype = np.dtype("O")


@pytest.mark.filterwarnings("ignore::zarr.core.dtype.common.UnstableSpecificationWarning")
@pytest.mark.parametrize("store", ["memory", "local"], indirect=["store"])
@pytest.mark.parametrize("dtype", numpy_str_dtypes)
@pytest.mark.parametrize("as_object_array", [False, True])
@pytest.mark.parametrize("compressor", [None, ZstdCodec()])
def test_vlen_string(
    store: Store, dtype: np.dtype[Any] | None, as_object_array: bool, compressor: Codec | None
) -> None:
    strings = ["hello", "world", "this", "is", "a", "test"]
    data = np.array(strings, dtype=dtype).reshape((2, 3))

    sp = StorePath(store, path="string")
    a = zarr.create_array(
        sp,
        shape=data.shape,
        chunks=data.shape,
        dtype=data.dtype,
        fill_value="",
        compressors=compressor,
    )
    assert isinstance(a.metadata, ArrayV3Metadata)  # needed for mypy

    # should also work if input array is an object array, provided we explicitly specified
    # a stringlike dtype when creating the Array
    if as_object_array:
        data_obj = data.astype("O")

        a[:, :] = data_obj
    else:
        a[:, :] = data
    assert np.array_equal(data, a[:, :])
    assert a.metadata.data_type == get_data_type_from_native_dtype(data.dtype)
    assert a.dtype == data.dtype

    # test round trip
    b = Array.open(sp)
    assert isinstance(b.metadata, ArrayV3Metadata)  # needed for mypy
    assert np.array_equal(data, b[:, :])
    assert b.metadata.data_type == get_data_type_from_native_dtype(data.dtype)
    assert a.dtype == data.dtype
